// Author: Pete Kirkham
//
// quad store matching query bench.
//
#include <xmmintrin.h>
#include <iostream>

#include "utils/core_utils.h"
#include "utils/string_utils.h"

#define BLOCK_SIZE (16*KiB)

bool linear_match_1         (unsigned int v0, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size) ;
bool bisection_match_1      (unsigned int v0, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size) ;

//~ bool linear_match1          (unsigned int v0, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size) ;
//~ __m128i* linear_match1_sse_2    (unsigned int v0, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size) ;
//~ __m128i* linear_match1_2        (unsigned int v0, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size) ;
//~ __m128i* bisection_match1_sse   (unsigned int v0, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size) ;
//~ __m128i* linear_match2_sse (unsigned int v0, unsigned int v1, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size) ;

void dump (__m128i* data, size_t limit) ;
void ndump (__m128i* data, size_t limit) ;


int main (int narg, char** argvp)
{
    size_t num_quads(1024);

    if (narg > 1)
    {
        num_quads = atoi(argvp[1]);
    }

    size_t num_passes(16);

    if (narg > 2)
    {
        num_passes = atoi(argvp[2]);
    }

    std::cout << "processing " << num_quads << " quads over " << num_passes << " passes." << std::endl;

    // determine method to use
    enum 
    {
        no_scan,
        linear,
        bisection,
    } scan_type(linear);


    for (int i(1); i < narg; ++i)
    {
        if (strcmp(argvp[i], "-no-scan") == 0)      scan_type = no_scan;
        if (strcmp(argvp[i], "-linear") == 0)       scan_type = linear;
        if (strcmp(argvp[i], "-bisection") == 0)    scan_type = bisection;
    }
    
    // create num_quads quads
    __m128i* data(0);
    
    posix_memalign(reinterpret_cast<void**>(&data),   16, num_quads*16);

    // fill in data
    for (size_t i(0); i < num_quads; ++i)
    {
        _mm_storeu_si128(data+i, _mm_set_epi32(1+i/35, 1+(i/7)%5, 1+i%7, 1));
    }

    // dump first 16 data
    //  dump(data, 16);

    __m128i* result_head(0);
    __m128i* result_tail(0);

    // scan for matches
    switch (scan_type)
    {
    case no_scan:
        break;
        
    case linear:
        for (size_t pass(0); pass < num_passes; ++pass)
        {
            linear_match_1(2, data, result_head, result_tail, num_quads);
        }
        break;
       
    case bisection:
        for (size_t pass(0); pass < num_passes; ++pass)
        {
            bisection_match_1(2, data, result_head, result_tail, num_quads);
        }
        break;
    };

    // output result
    size_t nresults(size_t(result_tail - result_head));

    if (nresults > 16)
    {
        std::cout << nresults << " results, first 16 printed: " << std::endl;
        dump(result_head, 16);
    }
    else
    {
        std::cout << nresults << " results: " << std::endl;
        dump(result_head, nresults);
    }
    
    free(data);

    std::cout << "done." << std::endl;
}

// matches the first value in quad using linear scan
bool linear_match_1      (unsigned int v0, __m128i* data, __m128i*& lo, __m128i*& hi, size_t size)
{
    __m128i mask(_mm_set_epi32(v0, 0, 0, 0));
    
    // scan for start
    for (size_t i(0); i < size; ++i)
    {
        __m128i in(_mm_loadu_si128(data+i));
        __m128i cmp(_mm_cmplt_epi32(in, mask));
        
//        std::cout << to_i4(in) << " ?= " << to_i4(mask) << " -> " << to_i4(cmp) << std::endl;
        
        if (!_mm_extract_epi16(cmp, 7))
        {
            lo = data + i;
            ++i;
            hi = data + i;
            
            // check last value
            {
                __m128i in(_mm_loadu_si128(data+size-1));
                __m128i cmp(_mm_cmpeq_epi32(in, mask));
                
                if (_mm_extract_epi16(cmp, 7))
                {
                    hi = data + size;
                    return true;
                }
            }
            
            for (; i < size; ++i)
            {
                __m128i in(_mm_loadu_si128(data+i));
                __m128i cmp(_mm_cmpgt_epi32(in, mask));
                
                if (_mm_extract_epi16(cmp, 7))
                {
                    hi = data + i;
                    break;
                }
            }
            
            return true;
        }
    }

    return false;
}

bool bisection_match_1 (unsigned int v0, __m128i* data, __m128i*& lo_in, __m128i*& hi_in, size_t size)
{
    __m128i mask(_mm_set_epi32(v0, 0, 0, 0));
    
    if (_mm_extract_epi16(_mm_cmpgt_epi32(_mm_loadu_si128(data), mask), 7)) return false;
    if (_mm_extract_epi16(_mm_cmplt_epi32(_mm_loadu_si128(data+size-1), mask), 7)) return false;

    size_t lo(0);
    size_t hi(size);
    
    if (_mm_extract_epi16(_mm_cmpeq_epi32(_mm_loadu_si128(data), mask), 7)) 
    {
        lo = 0;
    }
    else
    {
        size_t lo_lo(0);
        size_t lo_hi(size-1);
        // scan for start
        while (lo_lo != lo_hi)
        {
            size_t i(lo_lo + ((lo_hi-lo_lo)>>1));
            __m128i in(_mm_loadu_si128(data+i));
            __m128i cmp(_mm_cmplt_epi32(in, mask));

            //~ std::cout << '(' << lo_lo << ',' << lo_hi << ')'  << to_i4(in) << " ?= " << to_i4(mask) << " -> " << to_i4(cmp) << std::endl;
            
            if (_mm_extract_epi16(cmp, 7))
            {
                if (lo_lo == i) break;
                lo_lo = i;
            }
            else
            {
                lo_hi = i;
            }
        }
        
        lo = lo_hi;
    }

    if (_mm_extract_epi16(_mm_cmpeq_epi32(_mm_loadu_si128(data+size-1), mask), 7)) 
    {
        hi = size;
    }
    else
    {
        size_t hi_lo(lo);
        size_t hi_hi(size-1);

        // scan for start
        while (hi_lo != hi_hi)
        {
            size_t i(hi_lo + ((hi_hi-hi_lo)>>1));
            __m128i in(_mm_loadu_si128(data+i));
            __m128i cmp(_mm_cmpgt_epi32(in, mask));

            //~ std::cout << '(' << hi_lo << ',' << hi_hi << ',' << i << ')' << to_i4(in) << " ?= " << to_i4(mask) << " -> " << to_i4(cmp) << std::endl;
            
            if (_mm_extract_epi16(cmp, 7))
            {
                hi_hi = i;
            }
            else
            {
                if (hi_lo == i) break;
                hi_lo = i;
            }
        }
        
        hi = hi_hi;
    }

    if (lo != hi)
    {
        lo_in = data + lo;
        hi_in = data + hi;
        
        return true;
    }
    else
    {
        return false;
    }
}


// dumps the data
void dump (__m128i* data, size_t limit)
{
    for (size_t i(0); i < limit; ++i)
        std::cout << to_i4(_mm_loadu_si128(data+i)) << std::endl;
}

// dumps the data until first null
void ndump (__m128i* data, size_t limit)
{
    for (size_t i(0); i < limit; ++i)
        if (!(*((int*)(&data[i]))))
            break;
        else
            std::cout << to_i4(_mm_loadu_si128(data+i)) << std::endl;
}
